' FemtoBasic .01
#Compiler PBCC 6, PBWin 10
#Compile Exe
#Dim All
#Debug Error On
#Debug Display On

%PROLOGUE1 = 1
%PROLOGUE2 = 2
%EPILOGUE = 3
#Resource RcData, 1, "res\prologue1.txt"
#Resource RcData, 2, "res\prologue2.txt"
#Resource RcData, 3, "res\epilogue.txt"

$Title="FemtoBasic"
#If %Def( %PB_CC32 )
Macro MyPrint=Print
Macro MyInit=Con.Caption$=$Title
Macro MyInput=Line Input
Macro MyKey=WaitKey$
#Else
Macro MyPrint=Txt.Print
Macro MyInit=Txt.Window($Title, 10, 10 ) To hWin
Macro MyInput=Txt.Line.Input
Macro MyKey=Txt.WaitKey$
#EndIf

#Include "src\Bag.inc"
#Include "src\StringBag.inc"
#Include "src\StringLongMap.inc"

Global srcFile As String          ' Input file (somename.bas)
Global dstFile As String          ' Output file (somename.asm)
Global nextCharacter As Long      ' next input character
Global tokenValue As String       ' current token string
Global tokenType As Long          ' Type of token

' See Initialize for initial values
Global symbols As iSymbols        ' Variables, Subs, Functions, Classes, Interfaces, etc

#Include "src\BufferOutput.inc"
#Include "src\CodeGenAmd64.inc"
#Include "src\ConsoleInput.inc"
#Include "src\Errors.inc"
#Include "src\FileInput.inc"
#Include "src\FileOutput.inc"
#Include "src\Symbols.inc"

' See Initialize for initial values
Global operators As iStringBag:   ' values for valid operators
Global orOperator As iStringBag:  ' or/xor
Global andOperator As iStringBag: ' and
Global gen As iCodeGen:           ' code generator
Global codeOutput As iOutput:      ' output handler
Global keywords As iSymbols       ' Reserved keywords
Global rdr As iInput              ' program input

Function PBMain() As Long
	Local myCode As String

	' Main program
	Call Initialize()
	Call Block()
	Call gen.EndOfCode()

	' generate full code
	Local finalOutput As iOutput
	Call NewBufferOutput( finalOutput )

	gen.Output = finalOutput

	gen.EmitLine( "; " + srcFile )
	gen.Prologue1()
	gen.CreateVariables( symbols )
	gen.Prologue2()
	gen.IncludeLine( BufferOutput( codeOutput ) )
	gen.Epilogue()

	myCode = BufferOutput( finalOutput )

	' Save to disk
	Local wtr As iOutput
	Call NewFileOutput( wtr, dstFile )
	wtr.Write( myCode )
	FileOutputClose( wtr )
'  Local fh as Long
'  fh = freefile
'  open dstFile for output as #fh
'  Print #fh, myCode
'  Close #fh
End Function
Sub Initialize()
	srcFile = Command$( 1 )
	If srcFile = "" Then
		MyPrint "Usage: FemtoBasic srcFile"
		End
	End If
	dstFile = srcFile
	Local p As Long
	p  = InStr( -1, dstFile, "." )
	If p > 0 Then
		dstFile = Left$( dstFile, p ) + "asm"
	Else
		dstFile += ".asm"
	End If

	Call NewFileInput( rdr, srcFile )

	Call NewBufferOutput( codeOutput )
	Call NewCodeGenAmd64( gen, codeOutput )

	Call NewSymbols( symbols )

	Call NewSymbols( keywords )
	keywords.Add( "if", %SymbolType.IfSym )
	keywords.Add( "else", %SymbolType.ElseSym )
	keywords.Add( "endif", %SymbolType.EndIfSym )
	keywords.Add( "while", %SymbolType.WhileSym )
	keywords.Add( "wend", %SymbolType.WendSym )
	keywords.Add( "input", %SymbolType.InputSym )
	keywords.Add( "print", %SymbolType.PrintSym )
	keywords.Add( "printline", %SymbolType.PrintLineSym )
	keywords.Add( "dim", %SymbolType.DimSym )
	keywords.Add( "and", %SymbolType.AndSym )
	keywords.Add( "or", %SymbolType.OrSym )
	keywords.Add( "xor", %SymbolType.XorSym )
	keywords.Add( "not", %SymbolType.NotSym )

	Call NewStringBag( orOperator )
	orOperator.Add( "or" )
	orOperator.Add( "xor" )

	Call NewStringBag( andOperator )
	andOperator.Add( "and" )

	Call NewStringBag( operators )
	operators.Add( "," )
	operators.Add( ";" )
	operators.Add( "+" )
	operators.Add( "-" )
	operators.Add( "*" )
	operators.Add( "/" )
	operators.Add( ">" )
	operators.Add( ">=" )
	operators.Add( "<" )
	operators.Add( "<=" )
	operators.Add( "<>" )
	operators.Add( "=" )

	Call GetNextCharacter()
	Call GetNextToken()
End Sub
Sub GetNextCharacter()
	' Read new character from input stream
	nextCharacter = rdr.GetCharacter
End Sub
Sub GetNextToken()
	' Get the next input token
	Call SkipWhite()
	If nextCharacter = 0 Then
		tokenType = %SymbolType.EndProgramSym
		tokenValue = "!"
	ElseIf IsSymbolStart( nextCharacter ) Then
		Call GetSymbol()
	ElseIf IsDigitCharacter( nextCharacter ) Then
		Call GetNumber()
	Else
		Call GetOperator()
	End If
End Sub
Sub SkipWhite()
	' Skip over leading white space
	While IsWhitespaceCharacter( nextCharacter )
		Call GetNextCharacter()
	Wend
End Sub
Sub ChangeTypeIfIdentifierIsKeyword()
	' Scan the current identifier for keywords
	If tokenType = %SymbolType.IdentifierSym Then
		If keywords.Contains( tokenValue ) Then
			tokenType = keywords.Type( tokenValue )
		End If
	End If
End Sub
Sub GetSymbol()
	' Get an identifier
	Call SkipWhite()
	If Not IsSymbolCharacter( nextCharacter ) Then Call ExpectedError( "identifier" )
	tokenType = %SymbolType.IdentifierSym
	tokenValue = ""
	Do
		tokenValue += Chr$( nextCharacter )
		Call GetNextCharacter()
	Loop Until Not IsSymbolCharacter( nextCharacter )
	tokenValue = LCase$( tokenValue )
End Sub
Sub GetNumber()
	' Get a number
	Call SkipWhite()
	If Not IsDigitCharacter( nextCharacter ) Then Call ExpectedError( "number" )
	tokenType = %SymbolType.NumberSym
	tokenValue = ""
	Do
		tokenValue += Chr$( nextCharacter )
		Call GetNextCharacter()
	Loop Until Not IsDigitCharacter( nextCharacter )
End Sub
Sub GetOperator()
	' Get an operator
	Call SkipWhite()
	tokenValue = Chr$( nextCharacter )
	Call TestForOperator( tokenValue )
	tokenType = %SymbolType.OperatorSym
	Call GetNextCharacter()
	Select Case tokenValue
		Case ">"
			If nextCharacter = Asc( "=" ) Then
				tokenValue += Chr$( nextCharacter )
				Call GetNextCharacter()
			End If
		Case "<"
			If nextCharacter = Asc( ">" ) Or nextCharacter = Asc( "=" ) Then
				tokenValue += Chr$( nextCharacter )
				Call GetNextCharacter()
			End If
	End Select
End Sub
Function IsAlphabeticCharacter( c As Long ) As Long
	' Recognize an alpha character
	Select Case c
		Case Asc( "A" ) To Asc( "Z" ) : Function = -1: Exit Function
		Case Asc( "a" ) To Asc( "z" ) : Function = -1: Exit Function
	End Select
	Function = 0
End Function
Function IsDigitCharacter( c As Long ) As Long
	' Recognize a decimal digit
	Select Case c
		Case Asc( "0" ) To Asc( "9" ) : Function = -1: Exit Function
	End Select
	Function = 0
End Function
Function IsAlphaNumericCharacter( c As Long ) As Long
	' Recognize an alphanumeric character
	Function = IsAlphabeticCharacter( c ) Or IsDigitCharacter( c )
End Function
Function IsSymbolStart( c As Long ) As Long
	Function = c = Asc( "_" ) Or IsAlphabeticCharacter( c )
End Function
Function IsSymbolCharacter( c As Long ) As Long
	Function = c = Asc( "_" ) Or IsAlphaNumericCharacter( c )
End Function

Function IsAddOperatorCharacter( c As Long ) As Long
	' Recognize an addop
	Function = ( c = Asc( "+" ) Or c = Asc( "-" ) )
End Function
Function IsMultiplyOperatorCharacter( c As Long ) As Long
	' Recognize a mulop
	Function = ( c = Asc( "*" ) Or c = Asc( "/" ) )
End Function
Function IsOrOperatorCharacter( c As String ) As Long
	' Recognize a boolean orop
	Function = orOperator.Contains( c )
End Function
Function IsRelationOperatorCharacter( c As String ) As Long
	' Recognize a relop
	Function = ( c = "=" Or c = "<>" Or c = "<" Or c = ">" )
End Function
Function IsRelationOperator( c As String ) As Long
	Select Case c
		Case ">", ">=", "<", "<=", "<>", "="
			Function = -1
		Case Else
			Function = 0
	End Select
End Function

Function IsWhitespaceCharacter( c As Long ) As Long
	' Recognize white space
	Function = ( c = 32 Or c = 9 Or c = 13 Or c = 10 )
End Function
Sub TestForIdentifier()
	' Check to make sure the current token is an identifier
	If tokenType <> %SymbolType.IdentifierSym Then Call ExpectedError( "Identifier" )
End Sub
Sub TestForSymbolExists( symbol As String )
	' Check to see if an identifier is in the symbol table )
	' Report an error if it"s not.
	If Not symbols.Contains( symbol ) Then Call UndefinedError( symbol )
End Sub
Sub TestForDuplicateSymbol( identifier As String )
	' Check the symbol table for a duplicate identifier
	' Report an error if identifier is already in table. )
	If symbols.Contains( identifier ) Then Call DuplicateError( identifier )
End Sub
Sub TestForValue( match As String )
	' Match a specific input string
	If tokenValue <> match Then ExpectedError( "'" + match + "'" )
	Call GetNextToken()
End Sub
Sub TestForOperator( match As String )
	If Not operators.Contains( match ) Then Call InvalidOperator( match )
End Sub
Sub OperatorParenthesis()
	' Parse and translate a math factor
	If tokenValue = "(" Then
		Call GetNextToken()
		Call BooleanExpression()
		Call TestForValue( ")" )
	Else
		If tokenType = %SymbolType.IdentifierSym Then
			Call gen.LoadVar( tokenValue )
		ElseIf tokenType = %SymbolType.NumberSym Then
			Call gen.LoadConst( tokenValue )
		Else
			ExpectedError( "Math Factor" )
		End If
		Call GetNextToken()
	End If
End Sub

Sub OperatorMultiply()
	' Recognize and translate a multiply
	Call GetNextToken()
	Call OperatorParenthesis()
	Call gen.PopMul()
End Sub
Sub OperatorDivide()
	' Recognize and translate a divide
	Call GetNextToken()
	Call OperatorParenthesis()
	Call gen.PopDiv()
End Sub
Sub OperatorTerm()
	' Parse and translate a math term
	Call OperatorParenthesis()
	While IsMultiplyOperatorCharacter( Asc( tokenValue ) )
		Call gen.Push()
		Select Case tokenValue
			Case "*" : Call OperatorMultiply()
			Case "/" : Call OperatorDivide()
		End Select
	Wend
End Sub

Sub OperatorAdd()
	' Recognize and translate an add
	Call GetNextToken()
	Call OperatorTerm()
	Call gen.PopAdd()
End Sub
Sub OperatorSubtract()
	' Recognize and translate a subtract
	Call GetNextToken()
	Call OperatorTerm()
	Call gen.PopSub()
End Sub
Sub OperatorExpression()
	' Parse and translate an expression
	If IsAddOperatorCharacter( Asc( tokenValue ) ) Then
		Call gen.Clear()
	Else
		Call OperatorTerm()
	End If
	While IsAddOperatorCharacter( Asc( tokenValue ) )
		Call gen.Push()
		Select Case tokenValue
			Case "+" : Call OperatorAdd()
			Case "-" : Call OperatorSubtract()
		End Select
	Wend
End Sub
Sub CompareExpression()
	' Get another expression and compare
	Call GetNextToken()
	Call OperatorExpression()
	Call gen.PopCompare()
End Sub

Sub BooleanEqual()
	' Recognize and translate a relational "equals"
	Call CompareExpression()
	Call gen.SetEqual()
End Sub
Sub BooleanLessThanOrEqual()
	' Recognize and translate a relational "less than or equal"
	Call CompareExpression()
	Call gen.SetLessOrEqual()
End Sub
Sub BooleanNotEqual()
	' Recognize and translate a relational "not equals"
	Call CompareExpression()
	Call gen.SetNEqual()
End Sub
Sub BooleanLessThan()
	' Recognize and translate a relational "less than"
	Call CompareExpression()
	Call gen.SetLess()
End Sub
Sub BooleanGreaterThan()
	' Recognize and translate a relational "greater than"
	Call CompareExpression()
	Call gen.SetGreater()
End Sub
Sub BooleanGreaterThanOrEqual()
	' Recognize and translate a relational "greater than"
	Call CompareExpression()
	Call gen.SetGreaterOrEqual()
End Sub
Sub BooleanCompare()
	' Parse and translate a relation
	Call OperatorExpression()
	If IsRelationOperator( tokenValue ) Then
		Call gen.Push()
		Select Case tokenValue
			Case "=" : Call BooleanEqual()
			Case "<" : Call BooleanLessThan()
			Case ">" : Call BooleanGreaterThan()
			Case ">=" : Call BooleanGreaterThanOrEqual()
			Case "<=" : Call BooleanLessThanOrEqual()
			Case "<>" : Call BooleanNotEqual()
		End Select
	End If
End Sub

Sub BooleanNotFactor()
	' Parse and translate a boolean factor with leading not
	If tokenType = %SymbolType.NotSym Then
		Call GetNextToken()
		Call BooleanCompare()
		Call gen.NotPrimary()
	Else
		Call BooleanCompare()
	End If
End Sub
Sub BooleanTerm()
	' Parse and translate a boolean term
	Call BooleanNotFactor()
	While andOperator.Contains( tokenValue )
		Call gen.Push()
		Call GetNextToken()
		Call BooleanNotFactor()
		Call gen.PopAnd()
	Wend
End Sub

Sub BooleanOr()
	' Recognize and translate a boolean or
	Call GetNextToken()
	Call BooleanTerm()
	Call gen.PopOr()
End Sub
Sub BooleanXor()
	' Recognize and translate an exclusive or
	Call GetNextToken()
	Call BooleanTerm()
	Call gen.PopXor()
End Sub
Sub BooleanExpression()
	' Parse and translate a boolean expression
	Call BooleanTerm()
	While orOperator.Contains( tokenValue )
		Call gen.Push()
		Select Case LCase$( tokenValue )
			Case "or" : Call BooleanOr()
			Case "xor" : Call BooleanXor()
		End Select
	Wend
End Sub
Sub Assignment()
	' Parse and translate an assignment statement
	Local variable As String
	Call TestForSymbolExists( tokenValue )
	variable = tokenValue
	Call GetNextToken()
	Call TestForValue( "=" )
	Call BooleanExpression()
	Call gen.Store( variable )
End Sub
Sub DoIfStatement()
	' Recognize and translate an if construct
	Local label1, label2 As String
	Call GetNextToken()
	Call BooleanExpression()
	label1 = gen.NewLabel()
	label2 = label1
	Call gen.BranchFalse( label1 )
	Call Block()
	If tokenType = %SymbolType.ElseSym Then
		Call GetNextToken()
		label2 = gen.NewLabel()
		Call gen.Branch( label2 )
		Call gen.PostLabel( label1 )
		Call Block()
	End If
	Call gen.PostLabel( label2 )
	Call TestForValue( "endif" )
End Sub
Sub DoWhileStatement()
	' Parse and translate a while statement
	Local label1, label2 As String
	Call GetNextToken()
	label1 = gen.NewLabel()
	label2 = gen.NewLabel()
	Call gen.PostLabel( label1 )
	Call BooleanExpression()
	Call gen.BranchFalse( label2 )
	Call Block()
	Call TestForValue( "wend" )
	Call gen.Branch( label1 )
	Call gen.PostLabel( label2 )
End Sub
Sub InputToVariable()
	' Read a single variable
	Call TestForIdentifier()
	Call TestForSymbolExists( tokenValue )
	Call gen.EmitLine( "; INPUT " + tokenValue )
	Call gen.ReadConsole( tokenValue )
	Call GetNextToken()
End Sub
Sub DoInputStatement()
	' Process a read statement
	Call GetNextToken()
	Call InputToVariable()
	While tokenValue = ","
		Call GetNextToken()
		Call InputToVariable()
	Wend
End Sub
Sub DoPrintStatement()
	' Process a write statement
	Call GetNextToken()
	Call gen.EmitLine( "; PRINT" )
	If tokenValue = "," Then
		Call gen.WriteConsoleConst( "@tab" )
		Call GetNextToken()
	End If
	Call OperatorExpression()
	Call gen.WriteConsole()
	While tokenValue = "," Or tokenValue = ";"
		If tokenValue = "," Then
			Call gen.WriteConsoleConst( "@tab" )
		End If
		Call GetNextToken()
		Call OperatorExpression()
		Call gen.WriteConsole()
	Wend
End Sub
Sub DoPrintLineStatement()
	' Process a write statement
	Call GetNextToken()
	Call gen.EmitLine( "; PRINT" )
	If tokenValue = "," Then
		Call gen.WriteConsoleConst( "@tab" )
		Call GetNextToken()
	End If
	Call OperatorExpression()
	Call gen.WriteConsole()
	While tokenValue = "," Or tokenValue = ";"
		If tokenValue = "," Then
			Call gen.WriteConsoleConst( "@tab" )
		End If
		Call GetNextToken()
		Call OperatorExpression()
		Call gen.WriteConsole()
	Wend
	gen.WriteConsoleConst( "@crlf" )
End Sub
Sub CreateVariable()
	' Allocate storage for a variable
	Call GetNextToken()
	If tokenType <> %SymbolType.IdentifierSym Then ExpectedError( "Variable name" )
	Call TestForDuplicateSymbol( tokenValue )
	Call symbols.Add( tokenValue, %SymbolType.DimSym )
	'Call gen.Allocate( tokenValue, "0" )
	Call GetNextToken()
End Sub
Sub DoDimStatement()
	' Parse and translate global declarations
	Call CreateVariable()
	While tokenValue = ","
		Call CreateVariable()
	Wend
End Sub
Sub Block()
	' Parse and translate a block of statements
	Call ChangeTypeIfIdentifierIsKeyword()
	While Not ( tokenType = %SymbolType.EndIfSym Or tokenType = %SymbolType.WendSym Or tokenType = %SymbolType.ElseSym Or tokenType = %SymbolType.EndProgramSym )
		Select Case tokenType
			Case %SymbolType.DimSym : Call DoDimStatement()
			Case %SymbolType.IfSym : Call DoIfStatement()
			Case %SymbolType.WhileSym : Call DoWhileStatement()
			Case %SymbolType.InputSym : Call DoInputStatement()
			Case %SymbolType.PrintSym : Call DoPrintStatement()
			Case %SymbolType.PrintLineSym : Call DoPrintLineStatement()
			Case Else : Call Assignment()
		End Select
		Call ChangeTypeIfIdentifierIsKeyword()
	Wend
End Sub
