#Include This Once

#If Not %Def( %Const )
%Const = 1
Macro Const = Macro
Const LoadFactor = 0.75!
#EndIf

Sub NewBag( ByRef obj As iBag )
	obj = Class "cBag"
	If Not IsInterface( obj, iBag ) Then
		MyPrint "Failed to create bag"
		End
	End If
End Sub

Class cBag
	Instance bagEntries_() As iBagEntry
	Instance buckets_() As Long
	Instance count_ As Long
	Instance capacity_ As Long
	Instance full_ As Long

	Class Method Create()
		full_ = -1
		Me.Expand( 0 )
	End Method

	Class Method FindItem( ByRef pItem As iBagEntry ) As Long
			Local vBucketIndex As Long
			Local i As Long
			Local hc As Long
			Local be As iBagEntry

			hc = pItem.HashCode

			' Find entry
			vBucketIndex = hc Mod capacity_
			i = buckets_( vBucketIndex ) - 1
			While i >= 0 And IsObject( bagEntries_( i ) )
				be = bagEntries_( i )
				If be.HashCode = hc And pItem.Same( be ) Then
					Method = i
					Exit Method
				End If
				i = be.Next
			Wend

			Method = -1
	End Method

	' Adds a new string if it doesn't exist, either way returns the index of the item
	Class Method AddIfMissing( ByRef pItem As iBagEntry ) As Long
		Local vBucketIndex As Long
		Local i As Long

		i = Me.FindItem( pItem )
		If i>=0 Then
			Method = i
			Exit Method
		End If

		If count_ >= full_ Then
			Call Me.Expand( capacity_ + 1 )
		End If

		vBucketIndex = pItem.HashCode Mod capacity_

		pItem.Next = buckets_( vBucketIndex ) - 1
		bagEntries_( count_ ) = pItem
		Method = count_
		Incr count_
		buckets_( vBucketIndex ) = count_
	End Method

	' Expands the bag as needed
	Class Method Expand( ByVal capacity As Long )
		Local i As Long
		Local hc As Long
		Local vBucketIndex As Long
		Local pFull As Long


		capacity = BagGetCollectionSize( capacity )

		ReDim buckets_( capacity-1 )
		pFull = CLng( capacity * LoadFactor )
		ReDim Preserve bagEntries_( pFull )

		' Redistribute hashes
		For i = 0 To count_ - 1
			vBucketIndex = bagEntries_( i ).HashCode Mod capacity

			bagEntries_( i ).Next = buckets_( vBucketIndex ) - 1
			buckets_( vBucketIndex ) = i + 1
		Next
		capacity_ = capacity
		full_ = pFull
	End Method

	Interface iBag
		Inherit IUnknown

		' Returns index of added item
		Method Add( ByRef pItem As iBagEntry ) As Long
			Method = Me.AddIfMissing( pItem )
		End Method

		' Empty the bag
		Method Clear()
			Local i As Long
			For i=0 To count_-1
				bagEntries_( i ) = Nothing
			Next
			Reset buckets_()

			count_ = 0
		End Method

		' -1 if found, 0 if not found
		Property Get Contains( ByRef pItem As iBagEntry ) As Long
			Property = IIf&( Me.IndexOf( pItem ) <> -1, -1, 0 )
		End Property

		' Number of items in the bag
		Property Get Count() As Long
			Property = count_
		End Property


		' Index of item if found, -1 if not found
		Method IndexOf( ByRef pItem As iBagEntry ) As Long
			Method = Me.FindItem( pItem )
		End Method

		Method Remove( ByRef pItem As iBagEntry )
			Me.RemoveAt( Me.FindItem( pItem ) )
		End Method

		Method RemoveAt( ByVal i As Long )
			Local vHashCode As Long
			Local vBucketIndex As Long
			Local j As Long
			Local k As Long

			' This is generally super expensive, use with caution
			If i<0 Or i>= count_ Then Exit Method

			vHashCode = bagEntries_( i ).HashCode
			vBucketIndex = vHashCode Mod capacity_

			If buckets_( vBucketIndex ) - 1 = i Then
				buckets_( vBucketIndex ) = bagEntries_( i ).Next + 1: ' First item in list
			Else
				' Have prior entry's next point past the entry to remove
				j = buckets_( vBucketIndex ) - 1
				Do While j>=0
					If bagEntries_( j ).Next = i Then
						bagEntries_( j ).Next = bagEntries_( i ).Next
						Exit Loop
					End If
					j = bagEntries_( j ).Next
				Loop
			End If
			' Ok now remove bagEntries_( i )
			bagEntries_( i ) = Nothing
			' Shift rest of the objects back ( ouch exponential )
			For j = i To count_-2
				bagEntries_( j ) = bagEntries_( j+1 )
				' Fix up Next references
				If bagEntries_( j ).Next > i Then
					bagEntries_( j ).Next = bagEntries_( j ).Next - 1
				End If
			Next
			Decr count_
			' Release extra reference
			bagEntries_( count_ ) = Nothing
			' Adjust the bucket entries from i to the end, down by one
			If i > count_-1000 Then
				' This is faster if you're pretty close to the end, decent if you remove the last
				For j=i To count_-1
					vHashCode = bagEntries_( j ).HashCode
					vBucketIndex = vHashCode Mod capacity_
					If buckets_( vBucketIndex ) - 1 = j+1 Then
						Decr buckets_( vBucketIndex )
					End If
				Next
			Else
				' This is better if you are not close to the end
				For j=0 To capacity_
					If buckets_( j ) > i Then
						Decr buckets_( j )
					End If
				Next
			End If
			vHashCode = 0
		End Method

		' Get an item by index, returns empty string if out of bounds
		Property Get Item( ByVal index As Long ) As iBagEntry
			If index>=0 And index<count_ Then
				Property = bagEntries_( index )
			End If
		End Property
	End Interface
End Class

Function BagGetCollectionSize( ByVal capacity As Long ) As Long
	Static vPrimes() As Long
	Static vInit As Long
	Local i As Long

	If Not vInit Then
		ReDim vPrimes( 25 )
		Array Assign vPrimes() = _
						53, 97, 193, 389, 769, 1543, 3079, 6151, 12289 _
						, 24593, 49157, 98317, 196613, 393241, 786433 _
						, 1572869, 3145739, 6291469, 12582917, 25165843 _
						, 50331653, 100663319, 201326611, 402653189 _
						, 805306457, 1610612741
		vInit = -1
	End If

	For i=0 To 24
		If vPrimes( i )>capacity Then
			Function = vPrimes( i )
			Exit Function
		End If
	Next

	Function = vPrimes( 25 )
End Function
