#Include This Once

'Unique list of strings

Sub NewStringBag( obj As iStringBag )
	obj = Class "cStringBag"
	If Not IsInterface( obj, iStringBag ) Then
		MyPrint "Failed to create iStringBag"
		End
	End If
End Sub

Class cStringBag
	Instance bag_ As iBag
	Class Method Create()
		Call NewBag( bag_ )
	End Method

	Interface iStringBag
		Inherit IUnknown

		' Returns index of added item
		Method Add( ByRef pItem As String ) As Long
			Dim be As iBagEntry
			Call NewStringBagEntry( be, pItem )
			Method = bag_.Add( be )
		End Method

		' Empty the bag
		Method Clear()
			bag_.Clear()
		End Method

		' -1 if found, 0 if not found
		Property Get Contains( ByRef pItem As String ) As Long
			Dim be As iBagEntry
			Call NewStringBagEntry( be, pItem )

			Property = bag_.Contains( be )
		End Property

		' Number of items in the bag
		Property Get Count() As Long
			Property = bag_.Count
		End Property


		' Index of item if found, -1 if not found
		Method IndexOf( ByRef pItem As String ) As Long
			Dim be As iBagEntry
			Call NewStringBagEntry( be, pItem )
			Method = bag_.IndexOf( be )
		End Method

		Method Remove( ByRef pItem As String )
			Dim be As iBagEntry
			Call NewStringBagEntry( be, pItem )
			bag_.Remove( be )
		End Method

		Method RemoveAt( ByVal i As Long )
			bag_.RemoveAt( i )
		End Method

		' Get an item by index, returns empty string if out of bounds
		Property Get Item( ByVal index As Long ) As String
			Dim be As iBagEntry
			Dim iv As iStringValue
			Property = ""
			be = bag_.Item( index )
			iv = be
			If IsObject( iv ) Then
				Property = iv.Value
			End If
		End Property
	End Interface
End Class

Sub NewStringBagEntry( ByRef be As iBagEntry, ByRef value As String )
	Local b As iStringValue

	be = Class "cStringBagEntry"
	If Not IsInterface( be, iBagEntry ) Then
		MyPrint "Failed to create iBagEntry for " + value
		End
	End If
	b = be
	If Not IsInterface( b, iStringValue ) Then
		MyPrint "Failed to create iStringValue for " + value
		End
	End If
	b.Value = value
End Sub

'=================================================================
' StringBag entries ( internal implementation class )
'=================================================================
Class cStringBagEntry
	Instance hashCode_ As Long
	Instance next_ As Long
	Instance key_ As String

	Interface iStringValue
		Inherit IUnknown

		Property Get Value () As String
			Property = key_
		End Property
		Property Set Value ( ByVal pVal  As String )
			Local p As Byte Ptr
			Local i As Long

			If key_ <> pVal Then
				key_  = pVal

				hashCode_ = 1

				p = StrPtr( key_ )
				While @p
					hashCode_ = 1610612741 * hashCode_ + @p
					Incr p
				Wend
				hashCode_ And= &H7FFFFFFF
			End If
		End Property
	End Interface

	Interface iBagEntry
		Inherit IUnknown

		Method Same( ByVal value As iBagEntry ) As Long
			Local a As iStringValue
			Local b As iStringValue
			Local aa As String
			Local bb As String
			a = Me
			b = value
			If IsObject( a ) And IsObject( b ) Then
				aa = a.Value
				bb = b.Value
				Method = IIf&( LCase$( aa )=LCase$( bb ), 1, 0 )
			End If
		End Method

		Property Get HashCode () As Long
			Property = hashCode_
		End Property
		Property Set HashCode ( ByVal pHashCode  As Long )
			hashCode_  = pHashCode
		End Property

		Property Get Next () As Long
			Property = next_
		End Property
		Property Set Next ( ByVal pNext  As Long )
			next_  = pNext
		End Property
	End Interface
End Class
